LoadPackage("ace");
LoadPackage("polycyclic");
LoadPackage("kbmag");
LoadPackage("nq");


# Form the amalgam of the root groups.
#
# Returns the amalgam as a fp group, as well as a list homomorphisms
# from the root groups into the amalgam
AmalgamOfRootGroups := function(root_groups)
    local idx, i, j, k, n, tmp,
         rank2, IsoFpForSolvable,
         phi, H, F, rels, psi,
         AmalgamIdentifications,
         Utilde, hom, res, epi, gens;

    # Form all subsets of size 2 of the root groups
    idx := Combinations([1..Length(root_groups)], 2);
    n := Length(idx);

    # Form the rank 2 groups.
    rank2 := [];
    for i in idx do
        Add(rank2, Group(Union(List(root_groups{i}, GeneratorsOfGroup))));
    od;

    # Some (off by default) paranoia tests
    if AssertionLevel() >= 1 then
        # verify that the rank 2 groups intersect either trivially,
        # or in a root group
        for i in [1..n] do
            for j in [i+1..n] do
                tmp := Intersection(rank2[i], rank2[j]);
                k := Intersection(idx[i], idx[j]);
                if IsEmpty(k) then
                    Assert(1, IsTrivial(tmp));
                else
                    Assert(1, root_groups[k[1]] = tmp);
                fi;
            od;
        od;
    fi;

    # In order to form the amalgam, we first convert everything into fp groups.
    # Since we know that our groups are solvable (indeed, they are p-groups, so
    # they are even nilpotent), we may obtain finite presentation for them
    # efficiently by first converting to pc (i.e. polycyclic) groups.
    IsoFpForSolvable := function( G )
        local iso1, iso2;
        iso1 := IsomorphismPcGroup(G);
        iso2 := IsomorphismFpGroup(Image(iso1));
        return iso1*iso2;
    end;

    phi := List(rank2, IsoFpForSolvable);

    # Form the free product of the rank 2 subgroups; the amalgam
    # we seek is a quotient of that.
    H := FreeProduct(List(phi, Image));

    # The embeddings of our rank 2 groups into their free product H
    psi := List([1..n], i -> phi[i] * Embedding(H, i));

    # In order to form the amalgam, we need to factor out the
    # identifying relations given by their intersections.
    #
    # By construction (resp. assumption), we know that the intersection
    # of the rank 2 groups are precisely the root groups we started
    # with. The identiifcations we seek can be obtained by mapping (a
    # generating set of) each root group multiple times into H, once via
    # each embedding of the two rank 2 groups containing it. From this
    # it then is easy to derive the relators we need.
    #
    # The following function does just that, for a pair of rank 2 subgroups,
    # given by the embeddings psi1 and psi2.
    AmalgamIdentifications := function( grp, psi1, psi2 )
        local gens;
        gens := GeneratorsOfGroup(grp);
        return List(gens, g -> UnderlyingElement(ImageElm(psi1, g) / ImageElm(psi2,g)));
    end;

    # Form the actual amalgam: We need all relators of H, and add the
    # various identifying relators.
    F := FreeGroupOfFpGroup(H);
    rels := ShallowCopy(RelatorsOfFpGroup(H));
    for i in [1..Length(root_groups)] do
        k := PositionsProperty(idx, x -> i in x);
        k := Combinations(k, 2);
        for j in k do
            Append(rels, AmalgamIdentifications(root_groups[i], psi[j[1]], psi[j[2]]));
        od;
    od;
    Utilde := F / rels; # the amalgam we seek

    # Finally, compute the embeddings of the root groups into
    # the amalgam; these are useful for further computations.
    hom := GroupHomomorphismByImagesNC(H, Utilde);
    res := [];
    for i in [1..Length(root_groups)] do
        # Just pick any rank 2 group containing the i-th root group...
        j := PositionProperty(idx, x -> i in x);
        # ... then compose the embedding of that group into H with the
        #     quotient map from H onto Utilde ...
        epi := psi[j] * hom;
        # ... and finally create the embedding we seek from that.
        gens := GeneratorsOfGroup(root_groups[i]);
        Add(res, GroupHomomorphismByImagesNC(root_groups[i], Utilde, gens, List(gens, x -> x^epi)));
    od;

    return [Utilde, res];
end;

# Identify the unitriangular groups with a pcp group, fast. Using
# SubgroupUnitriangularPcpGroup is much more efficient than directly
# attempting to use IsomorphismPcGroup on the group U generated by all
# the root groups. That is so because the latter first tries to create a
# permutation group isomorphic to U, but that fails for large q (the
# permutation degree grows far too quickly).
RootGroups_to_PcpGroups := function(root_groups)
    local gens, F, B, U, gensU, i, gens_root;

    # Take all root group generators
    gens := Concatenation(List(root_groups, GeneratorsOfGroup));
    F := FieldOfMatrixList(gens);

    # SubgroupUnitriangularPcpGroup only accepts matrices over the prime field.
    # To work around that, we "blow up" the matrices if they are not.
    if not IsPrimeField(F) then
        B := Basis(F);
        gens := List(gens, g -> BlownUpMat(B, g));
        F := PrimeField(F);
    fi;

    # The matrices are lower triangular, SubgroupUnitriangularPcpGroup
    # expects upper triangular matrices. Luckily, the Cartan-Chevalley
    # involution x -> (x^-1)^T is a group automorphism of GL_n(q) that
    # maps one to the other.
    gens := List(gens, x->ImmutableMatrix(F, TransposedMat(x)^-1));
    U := SubgroupUnitriangularPcpGroup(gens);
    gensU := GeneratorsOfGroup(U);

    # We know that gens[i] maps to gensU[i] in U, so now we can the
    # original (matrix) root groups by corresponding subgroups of U.
    gens := Concatenation(List(root_groups, GeneratorsOfGroup));
    root_groups := ShallowCopy(root_groups);
    for i in [1..Length(root_groups)] do
        gens_root := GeneratorsOfGroup(root_groups[i]);
        gens_root := gensU{List(gens_root, g -> Position(gens, g))};
        root_groups[i] := Subgroup(U, gens_root);
    od;

    return root_groups;
end;

# Compute the covering degree of the amalgam Utilde of the
# rank 2 subgroups amalgamated along the root subgroups.
CoveringDegree_nq := function(U, root_groups)
    local ama, Utilde, iso, N, Unq,
          epi, H, F, rels, g, h, Utilde_ext, R, n, O, real_index, U12;

    Display("Computing amalgam of root groups");
    ama := AmalgamOfRootGroups(root_groups);
    Utilde := ama[1];

    # Compute the maximal nilpotent quotient of Utilde of class N,
    # where N is larger than the nilpotency class of U.
    #
    # If Utilde=U then this quotient will be isomorphic to U.
    #
    # Note: We could also use ANU pq instead of ANU nq.
    Display("Computing size of nilpotent quotient: ");
    if not IsEmpty(OptionsStack) then
        ResetOptionsStack(); # work around strange left overs on stack
    fi;
    N := NilpotencyClassOfGroup(U) + 5;
    epi := NqEpimorphismNilpotentQuotient(Utilde, N);
    Unq := Range(epi);

    Print("  Size(U)=", Size(U), ", Size(Unq)=", Size(Unq), ", covering degree at least ", Size(Unq)/Size(U), "\n");
    Print("    class U = ", NilpotencyClassOfGroup(U), "\n");
    Print("    class Unq = ", NilpotencyClassOfGroup(Unq), "\n");

    if Size(Unq) <> Size(U) then
        # If the sizes differ, then the covering degree is *at least* as large as the
        # following (but it could also be larger, in particular, infinite).
        return Size(Unq)/Size(U);
    fi;

    Print("Nilpotent quotient has right size; attempting Sims' trick to prove Utilde nilpotent\n");

    # Use the trick from [Sims94, 11.8]:
    # Introduce new generators in Utilde corresponding
    # to a pc series in Unq, then add relators which
    # define these in terms of the original generators.
    # Then apply Knuth Bendix

    H := FreeProduct(Utilde, Unq);
    F := FreeGroupOfFpGroup(H);
    rels := ShallowCopy(RelatorsOfFpGroup(H));
    # Add identifications
    for g in GeneratorsOfGroup(Unq) do
        # FIXME: We do not want PreImagesRepresentative here except for
        # the generators in the top layer; rather we want a "definition"
        # for the generators in terms of commutators of the higher-up
        # generators... I think... ?
        h := PreImagesRepresentative(epi, g);
        h := h^Embedding(H,1) / g^Embedding(H,2);
        Add(rels, UnderlyingElement(h));
    od;
    Utilde_ext := F / rels;

    # Use Sims' trick...
    R := KBMAGRewritingSystem(Utilde_ext);;
    n := Length(Alphabet(R));
    SetOrderingOfKBMAGRewritingSystem(R,"wreathprod", [n,n-1..1]);


    if true then
        # Using MAF as a plugin replacement for KBmag, we can just work straight forward
        SetInfoLevel(InfoRWS, 0);
        KnuthBendix(R);
    else
        SetInfoLevel(InfoRWS, 2);

        O:=OptionsRecordOfKBMAGRewritingSystem(R);
        #O.maxstoredlen:=[150,150];
        #O.maxstoredlen:=[300,300];
        O.maxstoredlen:=[600,600];
        #O.maxstoredlen:=[700,700];
        #O.maxstoredlen:=[800,800];
        KnuthBendix(R);
        Unbind(O.maxstoredlen);
        KnuthBendix(R);
    fi;

    if IsConfluent(R) then
        Print("Found confluent rewriting system! |Utilde| = ", Size(R),
                ", Utilde is nilpotent: ", Size(R)=Size(Unq), "\n");
        return Size(R) / Size(Unq);
    else
        Print("Failed to find confluent rewriting system\n");
        return fail;
    fi;
end;


# Compute the size of Utilde (and hence the covering degree) by
# computing the index of a large subgroup of known size.
CoveringDegree_index := function(U, root_groups)
    local ama, Utilde, gens, U12, U12tilde, iso, real_index, idx, mem;

    Display("Computing amalgam of root groups");
    ama := AmalgamOfRootGroups(root_groups);
    Utilde := ama[1];

    # The largest subgroup (= smallest index) whose size we know is U12.
    gens := [];
    Append(gens, GeneratorsOfGroup(root_groups[1]));  # generators of U1
    Append(gens, GeneratorsOfGroup(root_groups[2]));  # generators of U2
    U12 := Group( gens );

    gens := [];
    Append(gens, GeneratorsOfGroup(Image(ama[2][1])));  # generators of image of U1 in Utilde
    Append(gens, GeneratorsOfGroup(Image(ama[2][2])));  # generators of image of U2 in Utilde
    U12tilde := Subgroup(Utilde, gens);
    SetSize(U12tilde, Size(U12));

    # Simplify the presentations, carrying U12tilde along
    Display("Simplify presentations");
    iso := IsomorphismSimplifiedFpGroup(Utilde);
    Utilde := Image(iso);
    U12tilde := ImagesSet(iso, U12tilde);

    #
    Display("Computing index of U12 in Utilde...");

    idx := Size(U)/Size(U12);
    Print("  Expected index: ", idx, "\n");

    if false then
        TCENUM:=GAPTCENUM;
        SetInfoLevel(InfoFpGroup, 3);
        real_index := Index(Utilde, U12tilde);
    else
        #SetInfoACELevel(2);
        #SetInfoLevel(InfoFpGroup, 3);
        TCENUM := ACETCENUM;
        mem := (Length(GeneratorsOfGroup(Utilde)) + 1) * 2 * idx;
        real_index := Index(Utilde, U12tilde : max := 0, workspace := mem, messages := 10000, felsch:=0);
    fi;

    Print("  Actual index: ", real_index, "\n");
    Print("  Covering degree: ", real_index/idx, "\n");

    return real_index/idx;
end;


# Compute the covering degree by two different methods
CoveringDegree := function( unip )
    local root_groups, U, c1, c2;

    # First, transform the matrix groups into something we can
    # handle more easily.
    root_groups := RootGroups_to_PcpGroups( unip[2] );
    U := Parent(root_groups[1]);

    # Compute the covering degree via nilpotent quotient and Knuth-Bendix
    c1 := CoveringDegree_nq(U, root_groups);

    # Compute the covering degree via index of a large subgroup with known size
#     if c1 = 1 then # TODO: allow small powers of q?
#         c2 := CoveringDegree_index(U, root_groups);
#     else
         c2 := fail;
#     fi;

    return [c1, c2];
end;
